@@ -1,6 +1,6 @@
 #!/bin/sh
 #
-# Copyright (C) 2004-2005 Nokia Corporation.
+# Copyright (C) 2004-2006 Nokia Corporation.
 
 AF_PIDDIR=/tmp/af-piddir
 if [ ! -d $AF_PIDDIR ]; then
@@ -10,6 +10,11 @@ if [ ! -d $AF_PIDDIR ]; then
 	chmod 777 $AF_PIDDIR
 fi
 PIDFILE=$AF_PIDDIR/fb-progress.pid
+IMGDIR=/usr/share/icons/hicolor/scalable/hildon
+LOGO=qgn_indi_startup_nokia_logo.png
+BAR=qgn_indi_progressbar.png
+
+SECS=16
 
 case "$1" in
 start)	
@@ -18,7 +23,7 @@ start)
         if [ "x$BOOTREASON" != "xcharger" \
 	     -a ! -f /tmp/skip-fb-progress.tmp ]; then
         	echo "Starting: fb-progress"
-        	fb-progress -i 4 25 &
+        	fb-progress -l $IMGDIR/$LOGO -g $IMGDIR/$BAR $SECS &
         	echo "$!" > $PIDFILE
         	chmod 666 $PIDFILE
 	fi
